/*
 *                   jfjji,.           H A M M E R H E A D  S H A R K W A R E
 *                  ,#######EL:                  http://www.hhsw.de
 *                   .,j######Kj       
 *                    ;W########G:     This  code  is  free software; you can
 *                   j#####G ,D#L:     redistribute it and/or modify it under
 *             .,;;,G#####D.           the  terms  of  the GNU General Public
 *           jE##########G             License   as  published  by  the  Free
 *             .,W######D.             Software  Foundation; either version 2
 *             :L#####K##G             of  the  License,  or (at your option)
 *           .f#####G;.t##:            any later version.
 *         ;L###ELi.    ,D,            
 *       iEEGWj                        This  code  is distributed in the hope
 *      tW;                            that  it  will  be useful, but without
 *     jf                              any warranty; without even the implied
 *    .Kt                              warranty of merchantability or fitness
 *    i:                               for  a particular purpose. See the GNU
 *    ;                                General   Public   License   for  more
 *    .                                details (http://www.gnu.org).
 */
// use GNU extensions
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif // _GNU_SOURCE

#include <string.h>

#include "Raise.hh"

using namespace bbdock;

/**
 * Initialize object
 *
 * @param d - current display
 */
Raise::Raise( Display *d ) :
	display( d ),
	root( DefaultRootWindow( display ) )
{
}

/**
 * Activate a window wich has name in its title
 *
 * @param name - a substring of the title of a window
 */
bool Raise::raise( const char *name ) const
{
	Raise::Iterator	*i;

	if( !(i = getClientList()) )
		return false;

	for( Window w; (w = i->fetch()); )
	{
		char	*title;

		if( validateWindow( w ) &&
			(title = getWindowTitle( w )) &&
			strcasestr( title, name ) )
		{
			delete i;

			activateWindow( w );

			return true;
		}
	}

	delete i;
	
	return false;
}

/**
 * Intercept X errors
 */
static int throwError( Display *, XErrorEvent * )
{
	return False;
}

/**
 * Check if window is valid
 *
 * @param w - window id
 */
const bool Raise::validateWindow( Window w ) const
{
	XErrorHandler		handler;
	XWindowAttributes	attr;
	bool				valid = false;

	handler = XSetErrorHandler( throwError );
	if( XGetWindowAttributes( display, w, &attr ) )
		valid = true;
	XSetErrorHandler( handler );

	return valid;
}

/**
 * Read a window property
 *
 * @param w - window whose property to read
 * @param type - type of property
 * @param name - name of property
 * @param items - items retrieved (optional)
 */
unsigned char *Raise::getProperty( Window w, Atom type, const char *name, 
	unsigned long *items ) const
{
	Atom			returnedtype;
	int				format;
	unsigned long 	n;
	unsigned long	bytesafter;
	unsigned char	*prop;

	if( XGetWindowProperty( display, w, XInternAtom( display, name, False ), 
			0, 1024, False, type, &returnedtype, &format, &n, &bytesafter, 
			&prop ) != Success )
		return 0;

	if( returnedtype != type )
	{
		XFree( prop );
		return 0;
	}

	if( items )
		*items = n;

	return prop;
}

/**
 * Return title of a window
 *
 * @param w - window in question
 */
char *Raise::getWindowTitle( Window w ) const
{
	char *title;

	if( !(title = (char *) getProperty( w, XA_STRING, "WM_NAME" )) &&
		!(title = (char *) getProperty( w, XInternAtom( display, 
			"UTF8_STRING", False ), "_NET_WM_NAME" )) )
		return 0;

	return title; 
}

/**
 * Return list of all application windows
 */
Raise::Iterator *Raise::getClientList() const
{
	Window			*list;
	unsigned long	items;

	if( !(list = (Window *) getProperty( root, 
			XA_WINDOW, "_NET_CLIENT_LIST", &items )) &&
		!(list = (Window *) getProperty( root, 
			XA_CARDINAL, "_WIN_CLIENT_LIST", &items )) )
		return 0;

	if( !items )
		return 0;

	return new Raise::Iterator( list, items );
}

/**
 * Send a ClientMessage to some window
 *
 * @param w - target window
 * @param msg - message to send
 * @param data0 - data (optional)
 * @param data1 - data (optional)
 * @param data2 - data (optional)
 * @param data3 - data (optional)
 * @param data4 - data (optional)
 */
void Raise::sendClientMessage( Window w, const char *msg, 
	unsigned long data0, unsigned long data1, unsigned long data2, 
	unsigned long data3, unsigned long data4 ) const
{
	XEvent	event;
	long	mask = SubstructureRedirectMask | SubstructureNotifyMask;

	event.xclient.type = ClientMessage;
	event.xclient.serial = 0;
	event.xclient.send_event = True;
	event.xclient.message_type = XInternAtom( display, msg, False );
	event.xclient.window = w;
	event.xclient.format = 32;
	event.xclient.data.l[0] = data0;
	event.xclient.data.l[1] = data1;
	event.xclient.data.l[2] = data2;
	event.xclient.data.l[3] = data3;
	event.xclient.data.l[4] = data4;

	XSendEvent( display, root, False, mask, &event );
}

/**
 * Activate a window and switch to its workspace
 *
 * @param w - id of window to activate
 */
void Raise::activateWindow( Window w ) const
{
	unsigned long *desktop;

	if( (desktop = (unsigned long *) getProperty( w, XA_CARDINAL, 
			"_NET_WM_DESKTOP" )) ||
		(desktop = (unsigned long *) getProperty( w, XA_CARDINAL, 
			"_WIN_WORKSPACE" )) )
	{
		sendClientMessage( root, 
			"_NET_CURRENT_DESKTOP", *desktop );

		XFree( desktop );
	}

	sendClientMessage( w, "_NET_ACTIVE_WINDOW" );
	XMapRaised( display, w );
}
